/**
   \file
   \author Shin'ichiro Nakaoka
*/

#include "ForwardDynamics.h"
#include "Body.h"
#include "Link.h"
#include "Sensor.h"

using namespace cnoid;


ForwardDynamics::ForwardDynamics(BodyPtr body)
    : body(body)
{
    g.setZero();
    timeStep = 0.005;

    integrationMode = RUNGEKUTTA_METHOD;
    sensorsEnabled = false;
}


ForwardDynamics::~ForwardDynamics()
{

}


void ForwardDynamics::setTimeStep(double ts)
{
    timeStep = ts;
}


void ForwardDynamics::setGravityAcceleration(const Vector3& g)
{
    this->g = g;
}


void ForwardDynamics::setEulerMethod()
{
    integrationMode = EULER_METHOD;
}


void ForwardDynamics::setRungeKuttaMethod()
{
    integrationMode = RUNGEKUTTA_METHOD;
}


void ForwardDynamics::enableSensors(bool on)
{
    sensorsEnabled = on;
}


/// function from Murray, Li and Sastry p.42
void ForwardDynamics::SE3exp
(Vector3& out_p, Matrix3& out_R,
 const Vector3& p0, const Matrix3& R0, const Vector3& w, const Vector3& vo, double dt)
{
    double norm_w = w.norm();
	
    if(norm_w < std::numeric_limits<double>::epsilon()) {
        out_p = p0 + vo * dt;
        out_R = R0;
    } else {
        double th = norm_w * dt;
        Vector3 w_n = w / norm_w;
        Vector3 vo_n = vo / norm_w;
        Matrix3 rot(AngleAxisd(th, w_n));
		
        out_p.noalias() =
            rot * p0 + (Matrix3::Identity() - rot) * w_n.cross(vo_n) + (w_n * w_n.transpose()) * vo_n * th;

        out_R.noalias() = rot * R0;
    }
}


void ForwardDynamics::initializeSensors()
{
    body->clearSensorValues();

    if(sensorsEnabled){
        sensorCalculator.initialize(timeStep, g);
    }
}
